@@ -26,6 +26,15 @@ cv::MediaFrame::View cv::MediaFrame::access(Access code) const {
     return m->adapter->access(code);
 }
 
+cv::util::any cv::MediaFrame::blobParams() const
+{
+    return m->adapter->blobParams();
+}
+
+cv::MediaFrame::IAdapter* cv::MediaFrame::getAdapter() const {
+    return m->adapter.get();
+}
+
 cv::MediaFrame::View::View(Ptrs&& ptrs, Strides&& strs, Callback &&cb)
     : ptr   (std::move(ptrs))
     , stride(std::move(strs))
@@ -38,5 +47,11 @@ cv::MediaFrame::View::~View() {
     }
 }
 
+cv::util::any cv::MediaFrame::IAdapter::blobParams() const
+{
+    // Does nothing by default
+    return {};
+}
+
 cv::MediaFrame::IAdapter::~IAdapter() {
 }
